/*
 * Created on May 12, 2005
 */
package com.sunlabs.archsim.classRegistry;

import java.io.File;
import java.io.IOException;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;


import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.globals.Globals;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

import java.util.Stack;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 *
 * Read an XML file with the list of components that can be created by
 * component factories.
 *
 */
public class ClassRegistryLoader extends DefaultHandler {

	public static final String LIST= "component_list";
	public static final String COMPONENT= "component";
	public static final String TYPE= "type";
	public static final String CLASS= "class";
	public static final String TAG= "tag";
	public static final String NAME= "name";
	public static final String VALUE= "value";
	public static final String PARAMETER= "parameter";
	public static final String GLOBAL= "global";
	public static final String INPUT= "input";
	public static final String OUTPUT= "output";

	public static final String PAR_TYPE_DEFAULT= "int";
	
	
	
	Director director= Director.getInstance();
	
	/** registru where to register component classes */
	private ClassRegistry registry;
	
	/** the context of the current element */
	private Stack<String> context;
	
	/** accumulated charadted data */
	private String characterData= "";


	/** here we collect information about a component */
	private ComponentInfo componentInfo= null;
	
	/** Locator provided by the parser */
	private Locator locator;
	
	
	public ClassRegistryLoader() {
		registry= ClassRegistry.getInstance();
		context= new Stack<String>();
	} // ClassRegistryLoader constructor
	
	/** load components form a file */
	public ClassRegistry loadFile(String fileName) {
	       	SAXParserFactory factory = SAXParserFactory.newInstance();
	        factory.setValidating(true);
	        //factory.setNamespaceAware(true);
	        try {
	            // Parse the input
	            SAXParser saxParser = factory.newSAXParser();
	            
	            // set the buffer size for the reader to 8k
	            String bufferSizeProperty    = "http://apache.org/xml/properties/input-buffer-size";
	            Object bufferSize = new Integer(8192);
	            XMLReader reader= saxParser.getXMLReader();
	            reader.setProperty(bufferSizeProperty, bufferSize);
	            
	            saxParser.parse( new File(fileName), this);

	        } 
	        catch (SAXParseException spe) {
	           // Error generated by the parser
	           director.error("\n** Parsing error"
	              + ", line " + spe.getLineNumber()
	              + ", uri " + spe.getSystemId());
	           director.error("   " + spe.getMessage() );

	           // Use the contained exception, if any
	           Exception  x = spe;
	           if (spe.getException() != null)
	               x = spe.getException();
	           x.printStackTrace();
	           System.exit(1);

	        } catch (SAXException sxe) {
	            // Error generated by this application
	            // (or a parser-initialization error)
	      
	        		// first take a look at a message associated with
	        		// 
	        		String message= sxe.getMessage();
	        		if (!message.equals("")) {
	        			director.error(message);
	        		}
	        		
	        		Exception  x = sxe;
	        		if (sxe.getException() != null)
	            		x = sxe.getException();
	        		x.printStackTrace();
	        		System.exit(1);

	        } catch (ParserConfigurationException pce) {
	            	// Parser with specified options can't be built
	            	pce.printStackTrace();

	        } catch (IOException ioe) {
	        		// I/O error
	        		ioe.printStackTrace();
	        }
	        return registry;
	} // loadFile
	
	/** 
	 * The callback that provides a Locator - used to get line numbers
	 * for error messages, for example.
	 */
    public void setDocumentLocator(Locator l) {
    		locator = l;
    }

	
	/**
	 *  prep work to be done at the beginning of the document 
	 */
	public void startDocument() throws SAXException {
		// What do we do?
		// we have not entered the "root" element - list
		context.clear();
		
	} // startDocument
	
	/** 
	 * Finishing work to be done at the end of the XML file
	 */
	public void endDocument() throws SAXException {
		// What do we do?
		// nothing at the moment
	} // endDocument
	
	public void startElement (
			String namespaceURI,
			String lName, // local name
			String qName, // qualified name
			Attributes attrs) 
	throws SAXException {
		String eName = lName; // element name
	     if ("".equals(eName)) eName = qName; // namespaceAware = false
	     // establish the context
	     // note: the parser checks that the document is well-formed
	     // and valid
	     context.push(new String(eName));
	     
	     if (eName.equals(COMPONENT)) {
	    	 // prepare the new component info structure
	    	 componentInfo= new ComponentInfo();
	     }
	     else if (eName.equals(TAG)) {
	    	 // got a new tag, record it
	    	 if (attrs != null) {
	    		 for (int i=0; i< attrs.getLength(); i++) {
	    			 String attrName= attrs.getLocalName(i);
	    			 if (attrName.equals("")) attrName= attrs.getQName(i);
	    			 if (attrName.equals(VALUE)) {
	    				 componentInfo.addTag(new String(attrs.getValue(i).trim()));
	    			 }
	    		 }
	    	 }
	     }
	     else if (eName.equals(INPUT)) {
	    	 // record a new input
	    	 if (attrs != null) {
	    		 for (int i=0; i< attrs.getLength(); i++) {
	    			 String attrName= attrs.getLocalName(i);
	    			 if (attrName.equals("")) attrName= attrs.getQName(i);
	    			 if (attrName.equals(NAME)) {
	    				 componentInfo.addInput(new String(attrs.getValue(i).trim()));
	    			 }
	    		 }
	    	 }	    	 
	     }
	     else if (eName.equals(OUTPUT)) {
	    	 // record a new output
	    	 if (attrs != null) {
	    		 for (int i=0; i< attrs.getLength(); i++) {
	    			 String attrName= attrs.getLocalName(i);
	    			 if (attrName.equals("")) attrName= attrs.getQName(i);
	    			 if (attrName.equals(NAME)) {
	    				 componentInfo.addOutput(new String(attrs.getValue(i).trim()));
	    			 }
	    		 }
	    	 }	    	 	    	 
	     }
	     else if (eName.equals(PARAMETER)) {
	    	 int line= locator.getLineNumber();
	    	 if (attrs != null) {
    			 // look for parameter name and type
    			 String parName= null;
    			 String parType= PAR_TYPE_DEFAULT;
	    		 for (int i=0; i< attrs.getLength(); i++) {
	    			 String attrName= attrs.getLocalName(i);
	    			 if (attrName.equals("")) attrName= attrs.getQName(i);
	    			 if (attrName.equals(NAME)) {
	    				 parName= new String(attrs.getValue(i).trim());
	    			 }
	    			 else if (attrName.equals(TYPE)) {
	    				 parType= new String(attrs.getValue(i).trim());
	    			 }
	    		 } // for
	    		 if (parName != null && parType != null) {
	    			 componentInfo.addParameter(parName, parType);
	    		 }
	    		 else {
	    			 // DTD will insure we don't find ourselves here
	    			 // this code is here if we decide to ditch DTD
	    			 director.error("line: " + line 
	    					 + PARAMETER + " does not have both "
	    					 + NAME + " and " + TYPE + " defined. "
	    					 + "Values provided were: "
	    					 + NAME + " = " + parName
	    					 +" , " + TYPE + " = " + parType);
	    		 } // else
	    	 } // if !null	 	    	 
	     }
	     else if (eName.equals(GLOBAL)) {
	    	 int line= locator.getLineNumber();
	    	 if (attrs != null) {
    			 // look for parameter name and type
    			 String parName= null;
    			 String parType= PAR_TYPE_DEFAULT;
	    		 for (int i=0; i< attrs.getLength(); i++) {
	    			 String attrName= attrs.getLocalName(i);
	    			 if (attrName.equals("")) attrName= attrs.getQName(i);
	    			 if (attrName.equals(NAME)) {
	    				 parName= new String(attrs.getValue(i).trim());
	    			 }
	    			 else if (attrName.equals(TYPE)) {
	    				 parType= new String(attrs.getValue(i).trim());
	    			 }
	    		 } // for
	    		 if (parName != null && parType != null) {
	    			 componentInfo.addGlobal(parName, parType);
	    		 }
	    		 else {
	    			 // DTD will insure we don't find ourselves here
	    			 // this code is here if we decide to ditch DTD
	    			 director.error("line: " + line 
	    					 + GLOBAL + " does not have both "
	    					 + NAME + " and " + TYPE + " defined. "
	    					 + "Values provided were: "
	    					 + NAME + " = " + parName
	    					 +" , " + TYPE + " = " + parType);
	    		 } // else
	    	 } // if !null	 	    	 
	     }
	} // startElement
	
    public void endElement(String namespaceURI,
            String lName, // simple name
            String qName  // qualified name
           )
    throws SAXException {
    		
    	
    		// leave the current context
    		String currentContext= (String)context.pop();
    		if (currentContext.equals(TYPE)) {
    			// remember the name
    			componentInfo.setName(characterData);
    			// characterData= "";
    		}
    		else if (currentContext.equals(CLASS)) {

    			try {
    				// characterData= "";
    				componentInfo.setClass(Class.forName(characterData));
    			}
    			catch (ClassNotFoundException e) {
    				throw new SAXException("Error in line " 
    						+ locator.getLineNumber()
    						+ ": Class name \"" + characterData 
    						+ "\" not recognized", e);
    			}
    			if (!(Component.class.isAssignableFrom(componentInfo.getComponentType()))) {
    				// wrong type
    				throw new SAXException("Error in line " 
    						+ locator.getLineNumber()
    						+ ": Class " 
    						+ componentInfo.getComponentType().getName() 
    						+ " is not an instance of " 
    						+ Component.class.getName());
    			}
    			
    		}    		
    		else if (currentContext.equals(COMPONENT)) {
    			boolean fresh= registry.register(componentInfo.getName(), componentInfo);
    			if (!fresh) {
    				// check whether the tag has been used twice
    				Class c = registry.getComponentType(componentInfo.getName());
    				if (c != componentInfo.getComponentType()) { 
    					throw new SAXException("Error in line "
    						+ locator.getLineNumber()
    						+ ": Tag " + componentInfo.getName()
    						+ " has been used both for class " 
						+ c.getName() + " and class "
						+ componentInfo.getComponentType().getName());
    				}
    			} // if !fresh

    		}
    		
    		characterData="";
    		
    } // endElement
	
	public void characters(char buf[], int offset, int len)
	throws SAXException {
		// convert the character data into a string, trim the white space
		
		characterData+= (new String(buf, offset, len)).trim();		

	} // characters
	
    // treat validation errors as fatal
    public void error(SAXParseException e)
    throws SAXParseException
    {
        throw e;
    }

    // dump warnings too
    public void warning(SAXParseException err)
    throws SAXParseException
    {
        director.error("** Warning"
            + ", line " + err.getLineNumber()
            + ", uri " + err.getSystemId());
        director.error("   " + err.getMessage());
    }
	
	public static void main(String[] argv) {
		// a singleton
		ClassRegistry reg=  ClassRegistry.instance;
		
		Globals.getInstance().setParameter("global_log", "true");
		
		reg.load("library.xml");

		System.out.println(reg);
		
		boolean consistent= true;
		for (String name : reg.componentNames()) {
			ComponentInfo ci= reg.getComponentInfo(name);
			if (!ci.getName().equals(name)) {
				System.err.println("Component name mismatch: " 
						+ name + " != " + ci.getName());
				consistent= false;
			}
		}
		
		if (consistent) {
			System.out.println("Registry loader test passed");
		}
		else {
			System.out.println("Registry loader test failed");
		}
				
		
	} // main
	
} // ClassRegistryLoader



